!!! info ""

    A leckéhez tartozó teljes kód megtalálható a hozzá tartozó [branch-en](https://git.sch.bme.hu/kszk/devteam/vulkan-workshop/-/tree/lecture_02-final).

# Megjelenítés

## Felkészülés

Kezdés előtt ajánlott az [anyaghoz tartozó előadás](./Megjelenítés.pptx) átnézése.
Az előző leckének a kiegészített kódja megtalálható a `lecture_02-starter` branch-en.
Ez lesz a mostani kiinduló projektunk.

!!! note ""

    A [GLFW](https://www.glfw.org/) ablakozó könyvtárat fogjuk  használni, amihez előre elkészítettem egy minimális wrapper osztályt.

## *Extensions*

A Vulkan API *core* részét extension-ökkel szokták bővíteni.
Ezek a bővítmények nem garantáltak, hogy minden esetben elérhetőek/támogatottak; és mint minden használni kívánt dolgot, a kívánt extension-öket is **előre jelezni kell**.

Jelen alkalommal 2-féle bővítményre lesz szükségünk, hogy megjelenítéssel tudjunk foglalkozni.
Ezek a `VK_KHR_surface` és platform specifikus társai, illetve a `VK_KHR_swapchain`.

## `vk::SurfaceKHR`

A monitor és az ablak szorosan nem tartoznak a grafikus API-k hatásköre alá.
Vulkan-ban, lazán csatoltan, egy ablakozó könyvtár által készített ablak felületre tudunk rajzolni.
A kódunkban, hogy ezt a csatolást megteremtsük, vegyünk fel egy `const Window&` paramétert a `Renderer` konstruktorába.

```cpp title="Renderer.hpp"
class Renderer {
public:
    explicit Renderer(const Window& window);

private:
    // ...
    vk::UniqueSurfaceKHR m_surface;
    // ...
```

```cpp title="Renderer.cpp"
Renderer::Renderer(const Window& window)
    // ...
    m_surface{ window.create_vulkan_surface(m_instance.get()) },
    // ...
```

```cpp title="main.cpp"
Window   window{ 1'280, 720, "Vulkan workshop" };
Renderer renderer{ window };
```

Ha most futtatjuk az alkalmazást, egy felvillanó ablakot látunk.
Azonban egy csúnya hibaüzenet is megjelenik.

Engedélyezzük a felszínhez tartozó extension-öket, hogy minden simán menjen.

```cpp title="Renderer.cpp"
[[nodiscard]]
static auto create_instance() -> vk::UniqueInstance
    // ...
    const vk::InstanceCreateInfo create_info{
        // ...
        .enabledExtensionCount = static_cast<uint32_t>(Window::vulkan_instance_extensions().size()),
        .ppEnabledExtensionNames = Window::vulkan_instance_extensions().data(),
    };
    // ...
```

## Swapchain

A `vk::SurfaceKHR`-hez hasonlóan létezik egy `vk::SwapchainKHR` is, amivel "rajzolni" tudunk a kapott ablakfelszínre.
Az ablakfelszínnel ellentétben ezt mi fogjuk most létrehozni és felkonfigurálni.

??? example "Vegyük fel az új (még üres) `Swapchain` osztályt."

    ```cpp title="Swapchain.hpp"
    class Swapchain {};
    ```

Ahogyan majd később látni fogjuk, a swapchain nem mindig lesz valid állapotban.
Ez egy *immutable* objektum, és ha például az ablak mérete változik, akkor újra létre kell hozni.
Éppen ezért vegyünk fel egy **opcionális** swapchain változót a `Renderer`-be.

!!! example ""

    ```cpp title="Renderer.hpp"
    class Renderer {
        // ...
        std::optional<Swapchain> m_swapchain;
    };
    ```

### Extent

A swapchain feladata a [double/triple buffering](https://gameprogrammingpatterns.com/double-buffer.html) implementálása a rajzolt képek megjelenítéséhez, amihez értelemszerűen több képet is kezelnie kell.
A swapchain létrehozásakor legyen az első lépésünk ezen képek felbontásának megadása.

Fontos, hogy a létrehozott ablak felbontása nem feltétlenül ugyan az képernyő- és pixelkoordinátákban, ezért a *main.cpp*-ben használt `1280x720` nem fog működni.
A GLFW könyvtár a pixelkoordinátákban vett felszínt *framebuffer*-nek nevezi.
Ennek a méretét kell majd lekérni az ablaktól.

Sok esetben a kapott ablakfelszín segítségével lekérdezhető a hozzá tartozó méret is, de néha ez többféle érték is felhasználható.
Ilyenkor *clamp*-elni kell az általunk használni kívánt értéket a felszín által definiált min és max értékek közé.

??? example "Mindez így néz ki kódban"

    ```cpp title="Swapchain.hpp"
    class Swapchain {
    public:
        [[nodiscard]]
        static auto choose_extent(
            const vk::Extent2D&               framebuffer_size,
            const vk::SurfaceCapabilitiesKHR& surface_capabilities
        ) -> vk::Extent2D;
        // ...
    ```

    ```cpp title="Swapchain.cpp"
    auto Swapchain::choose_extent(
        const vk::Extent2D&               framebuffer_size,
        const vk::SurfaceCapabilitiesKHR& surface_capabilities
    ) -> vk::Extent2D
    {
        if (surface_capabilities.currentExtent.width != std::numeric_limits<uint32_t>::max())
        {
            return surface_capabilities.currentExtent;
        }

        return vk::Extent2D{
            .width = std::clamp(
                static_cast<uint32_t>(framebuffer_size.width),
                surface_capabilities.minImageExtent.width,
                surface_capabilities.maxImageExtent.width
            ),
            .height = std::clamp(
                static_cast<uint32_t>(framebuffer_size.height),
                surface_capabilities.minImageExtent.height,
                surface_capabilities.maxImageExtent.height
            ),
        };
    }
    ```

    !!! note ""

        A `surface_capabilities`-t majd az ablakfelszín segítségével fogjuk megkapni, a `framebuffer_size`-t pedig maga az ablak adja.

### Surface format

A következő lépés, hogy megadjuk a képek pixelformátumát.
Ez lehetne például a jól ismert RGBA (red, green, blue, alpha), ahol minden *color channel* 8 bit-et kap.
Azonban a legtöbb képernyő a BGRA-t támogatja jól.

A másik tulajdonság a színcsatornákban lévő értékek eloszlása/linearitása.
Mivel a szemünk nem lineáris intenzitással érzékeli az adott értékekhez tartozó fényt, ezért érdemes egy nem lineáris *color space* formátumot választani.
*(lásd. [gamma correction](https://en.wikipedia.org/wiki/Gamma_correction))*

??? example "Példa kód"

    ```cpp title="Swapchain.cpp"
    [[nodiscard]]
    static auto choose_swapchain_surface_format(
        const vk::SurfaceKHR     surface,
        const vk::PhysicalDevice physical_device
    ) -> vk::SurfaceFormatKHR
    {
        const auto available_surface_formats{ physical_device.getSurfaceFormatsKHR(surface) };

        for (const auto& surface_format : available_surface_formats) {
            if (surface_format.format == vk::Format::eB8G8R8A8Srgb
                && surface_format.colorSpace == vk::ColorSpaceKHR::eSrgbNonlinear)
            {
                return surface_format;
            }
        }

        for (const auto& surface_format : available_surface_formats) {
            if (surface_format.colorSpace == vk::ColorSpaceKHR::eSrgbNonlinear) {
                return surface_format;
            }
        }

        return available_surface_formats.front();
    }
    ```

### Képek száma

Ajánlott a minimum támogatott értéktől legalább eggyel nagyobb számú képet használni a swapchain-hez.
Illetve arra is figyelnünk kell, hogy ne lépjük át a maximumot.
(Ha a maximum-hoz definiált érték 0, akkor nincs maximum.)

??? example "Kódrészlet"

    ```cpp title="Swapchain.cpp"
    [[nodiscard]]
    static auto get_min_image_count(const vk::SurfaceCapabilitiesKHR& surface_capabilities
    ) noexcept -> uint32_t
    {
        uint32_t image_count = surface_capabilities.minImageCount + 1;
        if (surface_capabilities.maxImageCount > 0
            && image_count > surface_capabilities.maxImageCount)
        {
            image_count = surface_capabilities.maxImageCount;
        }

        return image_count;
    }
    ```

### Presentation mode

Még fontos kiemelni az [előadásdiákban](./Megjelenítés.pptx) is említett megjelenítési módokat (`Mailbox`, `Fifo`, ...).
Ahol nem annyira fontos az energiafogyasztás, ott erősen ajánlott a *Mailbox*.

??? example "Hogyan ellenőrizzük, hogy tudunk-e *Mailbox*-ot használni?"

    ```cpp
    [[nodiscard]]
    static auto choose_swapchain_present_mode(
        const vk::SurfaceKHR     surface,
        const vk::PhysicalDevice physical_device
    ) -> vk::PresentModeKHR
    {
        const auto present_modes{ physical_device.getSurfacePresentModesKHR(surface) };
        return std::ranges::find(present_modes, vk::PresentModeKHR::eMailbox)
                    != std::cend(present_modes)
                ? vk::PresentModeKHR::eMailbox
                : vk::PresentModeKHR::eFifo;
    }
    ```

A jelenlegi megoldásból kivettem ezt az opciót az egyszerűség kedvéért.
*Fifo*-t fogunk használni, mert az se rossz, és alapértelmezetten mindenhol támogatott.

### Létrehozás

A választott képformátumot és méretet tároljuk el, mert még később szükségünk lesz rá.

```cpp title="Swapchain.hpp"
class Swapchain {
public:
    // ...
    [[nodiscard]]
    auto get() const noexcept -> vk::SwapchainKHR;

    [[nodiscard]]
    auto extent() const noexcept -> vk::Extent2D;
    [[nodiscard]]
    auto format() const noexcept -> vk::Format;

private:
    vk::Extent2D           m_extent;
    vk::Format             m_format;
    vk::UniqueSwapchainKHR m_swapchain;

    explicit Swapchain(
        const vk::Extent2D&      extent,
        vk::Format               format,
        vk::UniqueSwapchainKHR&& swapchain
    ) noexcept;
};
```

```cpp title="Swapchain.cpp"
auto Swapchain::get() const noexcept -> vk::SwapchainKHR
{
    return m_swapchain.get();
}

auto Swapchain::extent() const noexcept -> vk::Extent2D
{
    return m_extent;
}

auto Swapchain::format() const noexcept -> vk::Format
{
    return m_format;
}

Swapchain::Swapchain(
    const vk::Extent2D&      extent,
    vk::Format               format,
    vk::UniqueSwapchainKHR&& swapchain
) noexcept
    : m_extent{ extent },
      m_format{ format },
      m_swapchain{ std::move(swapchain) }
{}
```

Végül hozzuk létre a swapchain-t.

```cpp title="Swapchain.hpp"
class Swapchain {
public:
    // ...
    [[nodiscard]]
    static auto create(
        vk::SurfaceKHR      surface,
        vk::PhysicalDevice  physical_device,
        uint32_t            queue_family_index,
        vk::Device          device,
        const vk::Extent2D& framebuffer_size
    ) -> std::optional<Swapchain>;
    // ...
```

0 szélességű/magasságú képet nem tudunk létrehozni, és nincs is értelme.
Ezért ebben az esetben visszaadunk egy `nullopt`-ot.
(Azaz az opcionális visszatérési értékben nem lesz semmi.)

```cpp title="Swapchain.cpp"
auto Swapchain::create(
    const vk::SurfaceKHR     surface,
    const vk::PhysicalDevice physical_device,
    const uint32_t           queue_family_index,
    const vk::Device         device,
    const vk::Extent2D&      framebuffer_size
) -> std::optional<Swapchain>
{
    const auto surface_capabilities{ physical_device.getSurfaceCapabilitiesKHR(surface) };

    const auto extent = choose_extent(framebuffer_size, surface_capabilities);
    if (extent.width == 0 || extent.height == 0) {
        return std::nullopt;
    }

    const auto surface_format{ choose_swapchain_surface_format(surface, physical_device) };

    auto swapchain{ create_swapchain(
        surface,
        queue_family_index,
        device,
        surface_capabilities,
        extent,
        surface_format
    ) };

    return Swapchain{ extent, surface_format.format, std::move(swapchain) };
}
```

```cpp title="Swapchain.cpp"
[[nodiscard]]
static auto create_swapchain(
    const vk::SurfaceKHR              surface,
    const uint32_t                    queue_family_index,
    const vk::Device                  device,
    const vk::SurfaceCapabilitiesKHR& surface_capabilities,
    const vk::Extent2D                extent,
    const vk::SurfaceFormatKHR        surface_format
) -> vk::UniqueSwapchainKHR
{
    const std::array queue_family_indices{ queue_family_index };

    const vk::SwapchainCreateInfoKHR create_info{
        .surface               = surface,
        .minImageCount         = get_min_image_count(surface_capabilities),
        .imageFormat           = surface_format.format,
        .imageColorSpace       = surface_format.colorSpace,
        .imageExtent           = extent,
        .imageArrayLayers      = 1,
        .imageUsage            = vk::ImageUsageFlagBits::eColorAttachment,
        .imageSharingMode      = vk::SharingMode::eExclusive,
        .queueFamilyIndexCount = static_cast<uint32_t>(queue_family_indices.size()),
        .pQueueFamilyIndices   = queue_family_indices.data(),
        .preTransform          = surface_capabilities.currentTransform,
        .compositeAlpha        = vk::CompositeAlphaFlagBitsKHR::eOpaque,
        .presentMode           = vk::PresentModeKHR::eFifo,
        .clipped               = vk::True
    };

    return device.createSwapchainKHRUnique(create_info);
}
```

??? note "Mi ez a sok paraméter amit megadtunk a swapchain-nek?"

    -   `imageArrayLayers` -
        Ez mindig egy, hacsak nem egy [sztereoszkopikus](https://en.wikipedia.org/wiki/Stereoscopy) alkalmazást fejlesztünk.

    -   `imageUsage` -
        Ezzel megadjuk, milyen módon fogjuk majd használni a képeket.
        Mivel jelenleg csak rajzolni fogunk rájuk, ezért *color attachment*-ként fogjuk őket használni, de akár *transfer*-hez is használhatnánk például *post-processing*-hez. 

    -   `imageSharingMode` -
        Több családból származó queue-k esetén beállíthatjuk, hogy egy kép melyik család tulajdona legyen.
        Mi csak csak egy queue-t használunk, ezért exkluzívan ő uralja a képet.
        Más esetben megadhatunk konkurrens tulajdon módot is.

    -   `preTransform`
        A képekhez hozzárendelhetünk valamilyen transformációt is.
        Amennyiben ezzel nem szeretnénk élni, úgy a `surface_capabilities.currentTransform`-ot kell használni.

    -   `compositeAlpha`
        Ezzel megadhatjuk, hogy átfedésben lévő ablakok hogyan hassanak a képre (blending).

    -   `clipped`
        Amennyiben egy másik ablak miatt fedésbe kerülnek bizonyos pixelek, eldönthetjük, hogy megtartsuk-e őket későbbi műveletektre.

```cpp title="Renderer.cpp"
Renderer::Renderer(const Window& window)
      // ...
      m_swapchain{ Swapchain::create(
          m_surface.get(),
          m_physical_device,
          m_queue_family_index,
          m_device.get(),
          window.framebuffer_size()
      ) }
{}
```

### Támogatás

A szemfüles olvasóknak feltűnhetett, hogy a `VK_KHR_SWAPCHAIN` bővítményt még nem engedélyeztük.
Illetve egy másik fontos dolog - hogyha használni is akarjuk a swapchain-t, akkor biztosra kell mennünk, hogy olyan queue-t választunk, ami támogat megjelenítést.
Gyorsan adjuk hozzá ezeket a setup-unkhoz.

!!! tip ""

    A következő kód részek másolása jóval könnyebb lesz a `lecture_02-final` branch-ből.
    Érdemes innentől inkább csak olvasni.

#### Enable extension

A `VK_KHR_SWAPCHAIN` egy *device extension*, tehát *device* szinten kell engedélyezni.
Ehhez első lépés, hogy olyan physical_device-t válasszunk, ami támogatja ezt.

??? example "Code snippets (*Renderer.cpp*)"

    ```cpp
    const std::vector<const char*> g_device_extensions{ VK_KHR_SWAPCHAIN_EXTENSION_NAME };
    ```

    ```cpp
    [[nodiscard]]
    static auto supports_required_extensions(const vk::PhysicalDevice physical_device) -> bool
    {
        const auto properties{ physical_device.enumerateDeviceExtensionProperties() };

        return std::ranges::all_of(g_device_extensions, [&](const char* extension_name) {
            return std::ranges::find_if(
                    properties,
                    [extension_name](const char* property_extension_name) {
                        return std::strcmp(extension_name, property_extension_name) == 0;
                    },
                    [](const vk::ExtensionProperties& extension_properties
                    ) -> const char* { return extension_properties.extensionName; }
                )
                != std::end(properties);
        });
    }
    ```

    ```cpp
    [[nodiscard]]
    static auto choose_physical_device(const vk::Instance instance) -> vk::PhysicalDevice
    {
        std::vector<vk::PhysicalDevice> physical_devices{};
        std::ranges::remove_copy_if(
            instance.enumeratePhysicalDevices(),
            std::back_inserter(physical_devices),
            [surface](vk::PhysicalDevice physical_device) {
                return !(
                    supports_required_extensions(physical_device)
                    && find_queue_family_index(physical_device).has_value()
                );
            }
        );
        // ...
    ```

    ```cpp
    [[nodiscard]]
    static auto create_device(const vk::PhysicalDevice physical_device) -> vk::UniqueDevice
    {
        // ...
        const vk::DeviceCreateInfo device_create_info{
            // ...
            .enabledExtensionCount   = static_cast<uint32_t>(g_device_extensions.size()),
            .ppEnabledExtensionNames = g_device_extensions.data(),
        };
    ```

#### Queue választás

??? example "Code snippets (*Renderer.cpp*)"

    ```cpp
    [[nodiscard]]
    static auto find_queue_family_index(
        const vk::PhysicalDevice physical_device,
        const vk::SurfaceKHR     surface
    ) -> std::optional<uint32_t>
    {
        uint32_t index{};
        for (const auto& properties : physical_device.getQueueFamilyProperties()) {
            if (properties.queueFlags & vk::QueueFlagBits::eGraphics
                && physical_device.getSurfaceSupportKHR(index, surface))
            {
                return index;
            }
            index++;
        }
        return std::nullopt;
    }
    ```

    ```cpp
    [[nodiscard]]
    static auto
        choose_physical_device(const vk::Instance instance, const vk::SurfaceKHR surface)
            -> vk::PhysicalDevice
    {
        std::vector<vk::PhysicalDevice> physical_devices{};
        std::ranges::remove_copy_if(
            instance.enumeratePhysicalDevices(),
            std::back_inserter(physical_devices),
            [surface](vk::PhysicalDevice physical_device) {
                return !(
                    supports_required_extensions(physical_device)
                    && find_queue_family_index(physical_device, surface).has_value()
                );
            }
        );
        // ...
    ```

    ```cpp
    [[nodiscard]]
    static auto
        create_device(const vk::PhysicalDevice physical_device, const vk::SurfaceKHR surface)
            -> vk::UniqueDevice
    {
        // ...
        const vk::DeviceQueueCreateInfo queue_create_info{
            .queueFamilyIndex = find_queue_family_index(physical_device, surface).value(),
            // ...
        };
    ```

    ```cpp
    Renderer::Renderer(const Window& window)
      : m_instance{ create_instance() },
        m_surface{ window.create_vulkan_surface(m_instance.get()) },
        m_physical_device{ choose_physical_device(m_instance.get(), m_surface.get()) },
        m_queue_family_index{
            find_queue_family_index(m_physical_device, m_surface.get()).value()
        },
        m_device{ create_device(m_physical_device, m_surface.get()) },
        m_queue{ m_device->getQueue(m_queue_family_index, 0) },
        m_swapchain{ Swapchain::create(
            m_surface.get(),
            m_physical_device,
            m_queue_family_index,
            m_device.get(),
            window.framebuffer_size()
        ) }
    {}
    ```
